package einsc.backend.opt;

import java.util.*;

import einsc.backend.Instruction;
import einsc.backend.cfg.*;

public class Liveness extends DataFlowAnalysis {
    public Liveness(CFG cfg) {
        super(cfg);
        analyseBackwards();
    }

    public BitSet liveIn(CFG.Node node) { return (BitSet)in(node); }
    public BitSet liveOut(CFG.Node node) { return (BitSet)out(node); }

    protected Object newValue() {
        return new BitSet();
    }

    protected Object cloneValue(Object v) {
        return ((BitSet)v).clone();
    }

    protected Object mergeValues(Object v1, Object v2) {
        BitSet res = new BitSet();
        res.or((BitSet)v1);
        res.or((BitSet)v2);
        return res;
    }

    /** Calculate transfer function for a BB. **/
    protected Transfer calculateTransfer(CFG.Node bb) {
        Instruction last = cfg.getLastInstruction(bb);
		if (last == null)
			return new LivenessTransfer(new BitSet(), new BitSet());
		
        BitSet kill = (BitSet)last.defs().clone();
        BitSet gen = (BitSet)last.uses().clone();

        ListIterator<Instruction> instrIt = bb.blockIterator(true);
        while (instrIt.hasPrevious()) {
            Instruction instr = /*(Instruction)*/instrIt.previous();
            gen.andNot(instr.defs());
            gen.or(instr.uses());
            kill.or(instr.defs());
        }
        return new LivenessTransfer(kill, gen);
    }

    protected class LivenessTransfer extends Transfer {
        public final BitSet defs, uses;

        public LivenessTransfer(BitSet kill, BitSet gen) {
            defs = kill;
            uses = gen;
        }

        public Object apply(Object out) {
            BitSet in = (BitSet) ((BitSet)out).clone();
            in.andNot(defs);
            in.or(uses);
            return in;
        }

        public String toString() {
            return "out:LiveVars -> (out - " + defs + ") U " + uses;
        }
    }

}
